@@ -5,7 +5,7 @@ use warnings;
 
 # ABSTRACT: Data store to support Net::LDAP::SimpleServer
 
-our $VERSION = '0.0.18';    # VERSION
+our $VERSION = '0.0.19';    # VERSION
 
 use 5.010;
 use Carp qw/carp croak/;
@@ -13,15 +13,22 @@ use UNIVERSAL::isa;
 use Scalar::Util qw(blessed reftype);
 
 use Net::LDAP::LDIF;
+use Net::LDAP::Util qw/canonical_dn/;
+
+use Net::LDAP::SimpleServer::Constant;
 
 sub new {
     my ( $class, $param ) = @_;
-    my $self = bless( { list => undef }, $class );
+    croak 'Must pass parameter!' unless defined($param);
 
-    croak 'Must pass parameter' unless $param;
+    # empty defaults
+    my $data = {
+        ldif_object => undef,
+        tree        => {},
+    };
 
+    my $self = bless( $data, $class );
     $self->load($param);
-
     return $self;
 }
 
@@ -30,30 +37,34 @@ sub load {
 
     croak 'Must pass parameter!' unless $param;
 
-    $self->{ldifobj} = _open_ldif($param);
-    $self->{list}    = _load_ldif( $self->{ldifobj} );
+    if ( blessed($param) && $param->isa('Net::LDAP::LDIF') ) {
+        $self->{ldif_object} = $param;
+    }
+    else {
+        $self->_open_ldif($param);
+    }
+    $self->_load_ldif();
     return;
 }
 
 sub ldif {
     my $self = shift;
-    return $self->{ldifobj};
+    return $self->{ldif_object};
 }
 
 #
 # opens a filename, a file-handle, or a Net::LDAP::LDIF object
 #
 sub _open_ldif {
+    my $self = shift;
     my $param = shift // '';
 
-    return $param if blessed($param) && $param->isa('Net::LDAP::LDIF');
-
     my $reftype = reftype($param) // '';
     if ( $reftype eq 'HASH' ) {
         croak q{Hash parameter must contain a "ldif" parameter}
           unless exists $param->{ldif};
 
-        return Net::LDAP::LDIF->new(
+        $self->{ldif_object} = Net::LDAP::LDIF->new(
             $param->{ldif},
             'r',
             (
@@ -62,21 +73,63 @@ sub _open_ldif {
                 : undef
             )
         );
+        return;
     }
 
     # Then, it must be a filename
-    croak q{Cannot find file "} . $param . q{"} unless -r $param;
+    croak q{Cannot read file "} . $param . q{"} unless -r $param;
 
-    return Net::LDAP::LDIF->new($param);
+    $self->{ldif_object} = Net::LDAP::LDIF->new($param);
+}
+
+sub _make_entry_path {
+    my $dn = shift;
+
+    $dn = $dn->dn() if $dn->isa('Net::LDAP::Entry');
+
+    return [ reverse( split( ',', canonical_dn($dn) ) ) ];
+}
+
+sub _make_entry {
+    my ( $entry, $tree, $current_dn, @path ) = @_;
+
+    $tree = {} unless defined($tree);
+    if ( scalar(@path) == 0 ) {
+        $tree->{_object} = $entry;
+    }
+    else {
+        my $next = $path[0];
+        $tree->{_object} = Net::LDAP::Entry->new($current_dn)
+          unless exists $tree->{_object};
+        $tree->{$next} = _make_entry(
+            $entry, $tree->{$next},
+            join( q{,}, $next, $current_dn ),
+            @path[ 1 .. $#path ]
+        );
+    }
+
+    return $tree;
+}
+
+sub _add {
+    my ( $self, $entry ) = @_;
+
+    my @path = @{ _make_entry_path($entry) };
+    my $tree = $self->{tree};
+    my $next = $path[0];
+    $tree->{$next} = _make_entry( $entry, $tree->{$next}, @path );
+
+    # line above is equivalent to
+    # _make_entry( $entry, $tree->{$next}, $next, @path[ 1 .. $#path ] );
 }
 
 #
 # loads a LDIF file
 #
 sub _load_ldif {
-    my $ldif = shift;
+    my $self = shift;
+    my $ldif = $self->{ldif_object};
 
-    my @list = ();
     while ( not $ldif->eof() ) {
         my $entry = $ldif->read_entry();
         if ( $ldif->error() ) {
@@ -85,20 +138,98 @@ sub _load_ldif {
             next;
         }
 
-        push @list, $entry;
+        $self->_add($entry);
     }
     $ldif->done();
-
-    my @sortedlist = sort { uc( $a->dn() ) cmp uc( $b->dn() ) } @list;
-
-    return \@sortedlist;
 }
 
-sub list {    ## no critic
-    return $_[0]->{list};
+sub _find_subtree {
+    my ( $tree, $rdn, @path ) = @_;
+
+    return unless exists $tree->{$rdn};
+    return $tree->{$rdn} if scalar(@path) == 0;
+    return _find_subtree( $tree->{$rdn}, @path );
 }
 
-1;            # Magic true value required at end of module
+sub find_tree {
+    my $self = shift;
+    my $dn   = shift;
+    $dn = $dn->dn() if $dn->isa('Net::LDAP::Entry');
+
+    return _find_subtree( $self->{tree}, @{ _make_entry_path($dn) } );
+}
+
+sub exists_dn {
+    my ( $self, $dn ) = @_;
+
+    my $tree = $self->find_tree($dn);
+    return defined($tree);
+}
+
+sub find_entry {
+    my ( $self, $dn ) = @_;
+
+    my $tree = $self->find_tree($dn);
+    return $tree->{_object} if defined($tree);
+    return;
+}
+
+sub _list {
+    my $tree = shift;
+
+    my @children_trees =
+      map { $tree->{$_} } ( grep { $_ ne '_object' } keys( %{$tree} ) );
+
+    return ( $tree->{_object}, ( map { ( _list($_) ) } @children_trees ) );
+}
+
+sub list {
+    my $self = shift;
+    my $tree = shift // $self->{tree}->{ ( keys( %{ $self->{tree} } ) )[0] };
+
+    return [ _list($tree) ];
+}
+
+sub _list_baseobj {
+    my $self  = shift;
+    my $dn    = shift;
+    my $entry = $self->find_entry($dn);
+
+    return unless defined($entry);
+
+    return [$entry];
+}
+
+sub _list_onelevel {
+    my $self = shift;
+    my $dn   = shift;
+    my $tree = $self->find_tree($dn);
+
+    return unless defined($tree);
+    my @children =
+      map { $tree->{$_}->{_object} }
+      ( grep { $_ ne '_object' } keys( %{$tree} ) );
+
+    return [ $tree->{_object}, @children ];
+}
+
+sub _list_subtree {
+    my $self = shift;
+    my $dn   = shift;
+    my $tree = $self->find_tree($dn);
+
+    return unless defined($tree);
+    return [ _list($tree) ];
+}
+
+sub list_with_dn_scope {
+    my ( $self, $dn, $scope ) = @_;
+
+    my @funcs = ( \&_list_baseobj, \&_list_onelevel, \&_list_subtree );
+    return $funcs[$scope]->( $self, $dn );
+}
+
+1;    # Magic true value required at end of module
 
 __END__
 
@@ -112,7 +243,7 @@ Net::LDAP::SimpleServer::LDIFStore - Data store to support Net::LDAP::SimpleServ
 
 =head1 VERSION
 
-version 0.0.18
+version 0.0.19
 
 =head1 SYNOPSIS
 
@@ -174,9 +305,34 @@ forms accepted by the constructor, except that it B<must> be specified.
 
 Returns the underlying C<< Net::LDAP::LDIF >> object.
 
-=head2 list
+=head2 find_tree( (DN|ENTRY) )
 
-Returns the list of entries.
+Search the store for a subtree, rooted on an entry that has the DN passed,
+or the DN of the C<< Net::LDAP::Entry >> object passed.
+Returns C<<undef>> if not found.
+
+B<WARNING:> this tree reflects the internal data structure of the store,
+and should not be used lightly. Whenever possible, use C<< find_entry() >>, C<exists_dn()>
+or C<< list() >>.
+
+=head2 find_entry( (DN|ENTRY) )
+
+Search the store for an entry, based on the DN passed,
+or the DN of the C<< Net::LDAP::Entry >> object passed.
+Returns C<<undef>> if not found.
+
+=head2 exists_dn( (DN|ENTRY) )
+
+Boolean version of C<<find_entry()>>
+
+=head2 list()
+
+Returns the list of C<< Net::LDAP::Entry >> objects in the store.
+
+=head2 list_with_dn_scope( DN, SCOPE )
+
+Returns a list of C<< Net::LDAP::Entry >> objects that conforms to the SCOPE
+applied to a DN, using standard LDAP rules.
 
 =head1 SEE ALSO
 
